/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.viewmodel

import android.app.Activity
import android.content.Context
import android.os.Bundle
import android.os.Parcelable
import android.os.ResultReceiver
import androidx.lifecycle.LiveData
import androidx.lifecycle.MutableLiveData
import androidx.lifecycle.ViewModel
import androidx.lifecycle.ViewModelProvider
import com.kakao.sdk.partner.friend.Constants
import com.kakao.sdk.partner.friend.Repository
import com.kakao.sdk.partner.friend.model.*
import com.kakao.sdk.partner.friend.network.model.PickerFriend
import com.kakao.sdk.partner.friend.util.FriendComparator
import com.kakao.sdk.partner.friend.util.PhonemeUtils
import com.kakao.sdk.partner.friend.util.StringUtils.getParsedString
import com.kakao.sdk.partner.friend.util.StringUtils.graphemeLength
import com.kakao.sdk.v2.partner.friend.R
import java.util.*
import kotlin.properties.Delegates

/**
 * @suppress
 */
internal class FriendViewModelFactory(
    private val resultReceiver: ResultReceiver,
    private val param: PickerRequestParams,
    private val isSingle: Boolean,
    private val isPopup: Boolean
) : ViewModelProvider.Factory {
    override fun <T : ViewModel?> create(modelClass: Class<T>): T {
        return if (modelClass.isAssignableFrom(FriendViewModel::class.java)) {
            FriendViewModel(resultReceiver, param, isSingle, isPopup) as T
        } else {
            throw IllegalArgumentException()
        }
    }
}

/**
 * @suppress
 */
internal class FriendViewModel(
    val resultReceiver: ResultReceiver,
    val param: PickerRequestParams,
    val isSingle: Boolean,
    val isPopup: Boolean
) : ViewModel() {
    /*
     LiveData는 MutableList.add 했을 때는 동작하지 않아서 MutableList와 LiveData를 따로 관리.
     MutableList.add() 후에 LiveData.value = MutableList 하는 방식으로 사용
     */
    private val totalList = mutableListOf<Item>() // 리스트에 보이는 전체 리스트
    private val pickedFriendsList = mutableListOf<Item.User>() // 선택된 친구 리스트
    private val favoritesList = mutableListOf<Item.User>() // 즐겨찾기 친구 리스

    private val _friendsList = mutableListOf<Item.User>() // 불러온 모든 친구 데이터가 저장된 리스트
    val friendsList: MutableList<Item.User> get() = _friendsList

    private val _pickerList = MutableLiveData<MutableList<Item>>()
    val pickerList: LiveData<MutableList<Item>> get() = _pickerList // 리스트에 보여지는 LiveData

    private val _pickedFriends = MutableLiveData<MutableList<Item.User>>()
    val pickedFriends: LiveData<MutableList<Item.User>> get() = _pickedFriends // 선택된 친구 LiveData

    private val _searchStatus = MutableLiveData<SearchStatus>().apply { value = SearchStatus.READY }
    val searchStatus: LiveData<SearchStatus> get() = _searchStatus

    private val _isError = MutableLiveData<Boolean>()
    val isError: LiveData<Boolean> get() = _isError

    private val repository = Repository

    private var pickerFriendOrder: PickerFriendOrder? = null
    private var showMyProfile by Delegates.notNull<Boolean>()
    private var showFavorite by Delegates.notNull<Boolean>()
    private var osFilter by Delegates.notNull<OsFilter>()
    private var disableSelectReasons: DisableSelectReason? = null
    private var displayAllProfile by Delegates.notNull<Boolean>()

    private val comparator: FriendComparator = FriendComparator() // 친구 리스트 정렬용 Comparator

    init {
        param.let {
            pickerFriendOrder = it.pickerFriendOrder
            showMyProfile = it.showMyProfile ?: Constants.DEFAULT_SHOW_MY_PROFILE
            showFavorite = it.showFavorite ?: Constants.DEFAULT_SHOW_FAVORITE
            osFilter = it.usingOsFilter ?: Constants.DEFAULT_OS_FILTER
            disableSelectReasons = it.disableSelectReasons
            displayAllProfile = it.displayAllProfile ?: Constants.DEFAULT_DISPLAY_ALL_PROFILE
        }
    }

    fun fetchData(context: Context) {
        if (pickerFriendOrder == null) {
            // pickerFriendOrder가 null이면 이름 순 정렬
            sortByNickname()
        } else {
            // null이 아니면 나이순 정렬 후 이름 순 정렬
            sortByAge()
        }

        if (showMyProfile) {
            totalList.add(Item.Header(context.getString(R.string.my_profile)))
            repository.me?.let {
                totalList.add(
                    Item.User(it.id, it.uuid, it.profileNickname, it.profileThumbnailImage)
                )
            }
            // 친구가 1명이라도 있으면 구분선을 그린다
            if (_friendsList.size > 0) {
                totalList.add(Item.Divider)
            }
        }

        // 친구가 한 명도 없는 경우
        if (_friendsList.size == 0) {
            _pickerList.value = totalList
            if (!showMyProfile) {
                _isError.value = true
            }
            return
        }

        if (showFavorite && favoritesList.size > 0) {
            totalList.add(Item.Header(context.getString(R.string.favorite)))
            totalList.addAll(favoritesList)
            totalList.add(Item.Divider)
        }
        totalList.add(Item.Header(context.getString(R.string.friend)))
        totalList.addAll(_friendsList)
        _pickerList.value = totalList
    }

    fun addCheckedFriend(user: Item.User) {
        pickedFriendsList.add(0, user)
        _pickedFriends.value = pickedFriendsList
        // 검색했을 때 즐겨찾기 친구는 뷰에서 사라지므로 검색 상태에서 친구를 선택하면 즐겨찾기 내에 있는 동일한 친구도 체크해줘야함
        updateFriend(user, true)
    }

    fun removeUncheckedFriend(user: Item.User) {
        pickedFriendsList.remove(user)
        _pickedFriends.value = pickedFriendsList
        // 검색시 뷰에 없는 선택된 친구를 제거하고, 해당 친구가 뷰에 다시 노출될 때 화면 동기화
        // (FriendsListAdapter에서 단일 리스트를 사용하기 때문에 어뎁터에서는 검색에 의해 뷰에 안보이는 선택된 친구를 제거할 수 없음)
        updateFriend(user, false)
    }

    fun searchFriends(context: Context, searchText: String) {
        if (searchText.isBlank()) {
            // 빈 검색창일 때
            _searchStatus.value = SearchStatus.DONE
            _searchStatus.value = SearchStatus.READY
            _pickerList.value = totalList // 전체 친구 리스트 보여줌
        } else {
            _searchStatus.value = SearchStatus.SEARCHING

            val filteredList =
                mutableListOf<Item>(Item.Header(context.getString(R.string.search_result)))

            // 검색 필터링
            filteredList.addAll(_friendsList.filter {
                PhonemeUtils.isMatch(it.nickname ?: "", searchText)
            })
            if (filteredList.size == 1) {
                filteredList.clear()
                filteredList.add(Item.EmptySearchResult)
            }
            _pickerList.value = filteredList // 필터링된 친구 보여줌
        }
    }

    fun sendResult() {
        val friends = mutableListOf<SelectedUser>()
        pickedFriendsList.forEach {
            if (repository.invitableProfilePermission && displayAllProfile) {
                friends.add(convertUserToResult(it, false))
            } else {
                friends.add(convertUserToResult(it, it.appRegistered != true))
            }
        }
        resultReceiver.send(
            Activity.RESULT_OK,
            Bundle().apply {
                putParcelableArrayList(
                    Constants.KEY_SELECTED_FRIENDS,
                    ArrayList<Parcelable>(friends)
                )
            }
        )
    }

    private fun convertFriendToUser(friend: PickerFriend) = Item.User(
        friend.id,
        friend.uuid,
        friend.profileNickname,
        friend.profileThumbnailImage,
        if (friend.allowedMsg == true) null else disableSelectReasons,
        friend.favorite,
        appRegistered = friend.appRegistered
    )

    private fun convertUserToResult(user: Item.User, enableMask: Boolean) = SelectedUser(
        if (user.appRegistered == true) user.id else null,
        user.uuid,
        if (enableMask) maskNickname(user.nickname ?: "") else user.nickname,
        if (enableMask) null else user.profileImageUrl,
        if (user.appRegistered == true) user.favorite else null
    )

    private fun sortByNickname() {
        repository.friends
            .filter {
                // os 필터링
                if (osFilter != OsFilter.ALL) it.talkOs == osFilter.name.toLowerCase(Locale.getDefault()) else true
            }.forEach { friend ->
                val user = convertFriendToUser(friend)
                if (friend.favorite == true) {
                    favoritesList.add(user)
                }
                _friendsList.add(user)
            }
        _friendsList.sortWith(comparator)
    }

    private fun sortByAge() {
        val filteredFriends = mutableListOf<Item.User>()
        val unfilteredFriends = mutableListOf<Item.User>()
        repository.friends
            .filter {
                // os 필터링
                if (osFilter != OsFilter.ALL) it.talkOs == osFilter.name.toLowerCase(Locale.getDefault()) else true
            }.forEach { friend ->
                val user = convertFriendToUser(friend)
                if (friend.ageFiltered == true) {
                    filteredFriends.add(user)
                } else {
                    unfilteredFriends.add(user)
                }
                if (friend.favorite == true) {
                    favoritesList.add(user)
                }
            }
        filteredFriends.sortWith(comparator)
        unfilteredFriends.sortWith(comparator)
        _friendsList.addAll(filteredFriends)
        _friendsList.addAll(unfilteredFriends)
    }

    private fun updateFriend(user: Item.User, isChecked: Boolean) {
        val indexOfFirst = totalList.indexOfFirst { it == user }
        val indexOfLast = totalList.indexOfLast { it == user }
        (totalList[indexOfFirst] as Item.User).isChecked = isChecked
        (totalList[indexOfLast] as Item.User).isChecked = isChecked
    }

    private fun maskNickname(nickname: String): String {
        return when (val length = nickname.graphemeLength()) {
            1 -> nickname
            3 -> {
                val list = nickname.getParsedString()
                "${list[0]}*${list[2]}"
            }
            else -> {
                val list = nickname.getParsedString()
                val s = StringBuilder()
                for (i in 0 until length) {
                    if (i < length / 2) {
                        s.append(list[i])
                    } else {
                        s.append("*")
                    }
                }
                s.toString()
            }
        }
    }
}